<?php

namespace Drupal\mapkit\Plugin\Mapkit\LocationInput;

use Drupal\Core\Cache\CacheableMetadata;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\Plugin\ContainerFactoryPluginInterface;
use Drupal\mapkit\Autocomplete\AutocompleteManagerInterface;
use Drupal\mapkit\Plugin\LocationResolverInterface;
use Symfony\Component\DependencyInjection\ContainerInterface;

/**
 * Input for entering lat/lng using the Mapkit autocomplete form element.
 *
 * @MapkitLocationInput(
 *   id = "autocomplete",
 *   label = @Translation("Location autocomplete"),
 * )
 */
class AutocompleteLocationInput extends TextfieldLocationInput implements ContainerFactoryPluginInterface {

  /**
   * The autocomplete strategy manager service.
   *
   * @var \Drupal\mapkit\Autocomplete\AutocompleteManagerInterface
   */
  protected AutocompleteManagerInterface $acManager;

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container, array $configuration, $plugin_id, $plugin_definition) {
    $instance = parent::create($container, $configuration, $plugin_id, $plugin_definition);

    $acManager = $container->get('strategy.manager.mapkit.autocomplete');
    $instance->setAutocompleteManager($acManager);

    return $instance;
  }

  /**
   * Set the autocomplete strategy manager service.
   *
   * @param \Drupal\mapkit\Autocomplete\AutocompleteManagerInterface $autocomplete_manager
   *   The autocomplete strategy manager service.
   */
  public function setAutocompleteManager(AutocompleteManagerInterface $autocomplete_manager): void {
    $this->acManager = $autocomplete_manager;
  }

  /**
   * {@inheritdoc}
   */
  public function defaultConfiguration() {
    $defaults = parent::defaultConfiguration();
    $defaults += [
      'handler' => NULL,
      'match_type' => 'address',
      'pattern' => NULL,
      'bound_type' => 'bias',
      'bounds' => [],
    ];

    return $defaults;
  }

  /**
   * {@inheritdoc}
   */
  public function buildInputElements($values, array &$elements, FormStateInterface $form_state, ?LocationResolverInterface $resolver = NULL): void {
    $config = $this->getConfiguration();
    $titleDisplay = $config['label_display'] ?? 'before';

    // Create and update caching dependencies for this form input when using
    // a location resolver.
    $cache = CacheableMetadata::createFromRenderArray($elements);
    if ($resolver) {
      $cache->addCacheableDependency($resolver);

      // If missing a default value, see if one can be populated.
      if (empty($values['loc']['text']) && empty($values['loc']['latlng'])) {
        $values['loc']['text'] = $resolver->getLocationText();

        if ($latLng = $resolver->getLatLng()) {
          $values['loc']['latlng'] = $latLng['lat'] . ',' . $latLng['lng'];
        }
      }
    }

    $elements['#tree'] = TRUE;

    if ($bounds = $config['bounds']) {
      $bounds['type'] = $config['bound_type'];
    }

    $elements['loc'] = [
      '#type' => 'mapkit_autocomplete',
      '#title' => $config['label'],
      '#title_display' => $titleDisplay,
      '#default_value' => $values['loc'] ?? NULL,
      '#data_keys' => ['latlng'],
      '#handler' => $config['handler'],
      '#bounds' => $bounds,
      '#device_location' => $config['device_location'] ?? FALSE,
      '#filter_pattern' => $config['pattern'],
      '#placeholder' => $config['placeholder'],
      '#autocomplete_match_type' => $config['match_type'],
    ];

    $radii = [];
    foreach ($config['radii'] as $radius) {
      $radiusValue = (string) $radius['value'];
      $radii[$radiusValue] = $radius['label'];
    }
    $elements['radius'] = [
      '#type' => 'select',
      '#title' => $config['radius_label'] ?? '',
      '#title_display' => $titleDisplay,
      '#options' => $radii,
      '#default_value' => $values['radius'] ?? NULL,
    ];

  }

  /**
   * {@inheritdoc}
   */
  public function buildConfigurationForm(array $form, FormStateInterface $form_state) {
    $form = parent::buildConfigurationForm($form, $form_state);

    $config = $this->getConfiguration();
    $handlerOpts = $this->acManager->getAutocompleteOptions();
    if (1 === count($handlerOpts)) {
      $config['handler'] = key($handlerOpts);
    }

    $form['handler'] = [
      '#type' => 'select',
      '#title' => $this->t('Autocomplete handler'),
      '#required' => TRUE,
      '#options' => $handlerOpts,
      '#default_value' => $config['handler'],
      '#weight' => -5,
    ];
    $form['match_type'] = [
      '#type' => 'select',
      '#title' => $this->t('Result types'),
      '#options' => [
        '' => $this->t('Any'),
        'address' => $this->t('Address level results'),
        'city' => $this->t('Cities or Postal code'),
        'region' => $this->t('City, State or Postal code'),
      ],
      '#default_value' => $config['match_type'],
      '#weight' => -4,
    ];
    $form['pattern'] = [
      '#type' => 'textfield',
      '#title' => $this->t('Results pattern (regular expression)'),
      '#default_value' => $config['pattern'],
      '#description' => $this->t('Used to filter which suggestions get displayed in the autocomplete. Leave empty to show all results from the autocomplete service.'),
    ];

    $form['bound_settings'] = [
      '#type' => 'details',
      '#title' => $this->t('Search bounds'),
      '#description_display' => 'before',
      '#description' => $this->t('Provide a map bounds (in latitude/longitude ranges) to bias results. Results are more likely to appear from these bounds, though the specific implemention and rules will be up to the autocomplete service used.'),
      '#weight' => 20,

      'bound_type' => [
        '#type' => 'select',
        '#title' => $this->t('Bounds type'),
        '#parents' => [...$form['#parents'], 'bound_type'],
        '#options' => [
          'bias' => $this->t('Bias'),
          'restriction' => $this->t('Restriction'),
        ],
        '#default_value' => $config['bound_type'] ?? NULL,
      ],
      'bounds' => [
        '#type' => 'latlng_bound',
        '#title' => $this->t('Bounds latitude/longitude'),
        '#parents' => [...$form['#parents'], 'bound'],
        '#default_value' => $config['bounds'] ?? [],
      ],
    ];

    return $form;
  }

  /**
   * {@inheritdoc}
   */
  public function submitConfigurationForm(array &$form, FormStateInterface $form_state) {
    parent::submitConfigurationForm($form, $form_state);

    $this->configuration['handler'] = $form_state->getValue('handler');
    $this->configuration['match_type'] = $form_state->getValue('match_type');
    $this->configuration['pattern'] = $form_state->getValue('pattern');
    $this->configuration['bound_type'] = $form_state->getValue('bound_type');
    $this->configuration['bounds'] = $form_state->getValue('bound');
  }

}
